from typing import Callable, TypeVar, Union

K = TypeVar('K')
T = TypeVar('T')
def Switch(arg: K, options: dict[K, Callable[[], T]], default: Callable[[], T] = lambda: None) -> Union[T, None]:
    enter = False
    for key, fn in options.items():
        if arg == key:
            return fn()
    return default()



import datetime, time
from . import config

def current_milli_time() -> int:
    return round(time.time() * 1000)

def current_sec_time() -> float:
    return current_milli_time() / 1000

def current_time() -> float:
    return time.time()

def format_from_utc(ts) -> str:
    return datetime.datetime.fromtimestamp(ts, tz=config.TIME_ZONE).strftime("%Y-%b-%d %I:%M:%S %p")


import requests as Requests
from typing import Callable, TypeVar, Union
import json

from limes_common import config
from limes_common.connections import Connection
from limes_common.models import Primitive, Model, provider as Models
from limes_common.models.http import GET, POST

T = TypeVar('T')
class HttpConnection(Connection):
    def __init__(self, url) -> None:
        super().__init__()
        self.session = Requests.session()
        self._URL: str = url
        self.methods: dict[str, Callable[..., Requests.Response]] = {
            GET: self.session.get,
            POST: self.session.post
        }

    def _makeUrl(self, ep: str):
        return (self._URL if self._URL.endswith('/') else self._URL[:-1]) + ep

    def _makeJson(self, data: Models.ProviderRequest) -> dict:
        d = data.ToDict()
        return d

    def _makeHeader(self):
        return {}

    def _makeParams(self, req: Models.ProviderRequest):
        return req.QueryParams

    def _makeParseRequest(self, req: Models.ProviderRequest, parser: Callable[[dict, Models.ProviderResponse], T], default: T) -> T:
        raw = self.MakeRequest(req)
        if isinstance(raw.Body, dict):
            return parser(raw.Body, raw)
        else:
            return default

    def GetSchema(self) -> Models.Schema:
        s = Models.Schema()
        s.Code = 503
        s.Error = 'for static http providers, please refer to their specific documentation'
        return s

    def MakeRequest(self, request: Models.ProviderRequest) -> Models.GenericResponse:
        ep = request.TargetEndpoint
        doRequest = self.methods.get(request.Method, None)
        if ep is None or doRequest is None:
            msg = 'no endpoint given' if ep is None else 'bad http method [%s]'%request.Method
            return Models.GenericResponse({}, 0, msg)
        
        if isinstance(request, Models.GenericRequest):
            if isinstance(request.Body, dict):
                body = request.Body
            else:
                body = {'body': request.Body}
        else:
            body = self._makeJson(request)
        res = doRequest(
            self._makeUrl(ep),
            json=body,
            headers = self._makeHeader(),
            params = self._makeParams(request),
            timeout = config.HTTP_TIMEOUT
        )
        code = res.status_code
        try:
            body: dict[str, Primitive] = json.loads(res.text)
        except Exception:
            body = {'raw': res.text}
        return Models.GenericResponse(body, code)


from pytz import timezone

_DEBUG = False

TIME_ZONE = timezone('Canada/Pacific')
ENCODING = 'utf-8'
if _DEBUG:
    HTTP_TIMEOUT = 10
else:
    HTTP_TIMEOUT = 5

CSRF_NAME = 'X-CSRFToken'

# ELAB_URL = 'https://us.elabjournal.com/'
ELAB_URL = 'https://elab.msl.ubc.ca/'
ELAB_API = '%sapi/v1/' % ELAB_URL

SERVER_API_VER = 'd1'
if _DEBUG:
    SERVER_BIND = '127.0.0.1:8001'
else:
    SERVER_BIND = 'sh-lims.microbiology.ubc.ca:8001'
SERVER_URL = 'http://%s/api/%s/' % (SERVER_BIND, SERVER_API_VER)
SERVER_DB_PATH = 'limes_server/db'

# PROVIDER_STATICS_PATH = 'limes_core/example_static_providers.json'
PROVIDER_STATICS_PATH = 'limes_core/static_providers.json'
PROVIDER_DEFAULT_TRANSACTION_TIMEOUT = 5
PROVIDER_DEFAULT_CONNECTION_TIMEOUT = 60

ELAB_CACHE = 'elab_cache.npy'


from .limes import Limes

import sys
from typing import List
from getpass import getpass
from functools import reduce

from .tools.qol import Switch, T
from .res import strings as Constants
from limes_common import config

def main(args:List[str] = sys.argv[1:]):
    # for later abstraction of help
    class Command:
        def __init__(self, schemas, options) -> None:
            self.schemas = schemas
            self.options = options

        def printHelp(self):
            self.printLine()
            print('Required:')

        def printLine(self):
            print('====================================================================')

    def Login():
        userKeys = ['-u', '--username']
        passKeys = ['-p', '--password']

        tries = 3
        user = input('Username: ')
        for _ in range(tries):
            try:
                password = getpass()
            except KeyboardInterrupt:
                return

            if limes.Login(user, password):
                return
        return '%s failed attempts, check your connection and credentials' % tries

    def _printHelp(unkownArg=False):
        msg = 'use [-h] or [--help] to view options\nor view available documentation at https://github.com/Tony-xy-Liu/Limes/'
        if unkownArg:
            msg = '[%s] was not recognized\n%s' % (unkownArg, msg)
        return msg

    def _getVal(keys: List[str], args: List[str]) -> tuple[bool, str]:
        next = 1
        for item in args:
            if item in keys and next < len(args):
                return True, args[next]
            next +=1
        return False, ""

    limes = Limes()
    arg = args[0] if len(args) > 0 else ''

    def base():
        print(Constants.version)
        limes._auth()

    # the switch doesn't need to be fast, make it easier to make
    # also need to add hints here
    return Switch(arg, {
        '': base,
        '-h': lambda: Constants.help,
        '--help': lambda: Constants.help,
        'login': Login
        # 'search': Search,
        # 'blast': Blast,
        # 'add': Add,
    }, lambda: _printHelp(arg))

if __name__ == "__main__":
    args = sys.argv[1:]
    a = args[0] if len(args) > 0 else ''
    try:
        end = main()
    except KeyboardInterrupt:
        print('KeyboardInterrupt')
        end = None
    if not end is None and end != '':
        sys.exit(end)
    else:
        sys.exit()


from typing import Any, Union
import os

from limes.tools.qol import Switch

from limes_common.connections.statics import eLab, server
from limes_common.models import Primitive, Model, server as ServerModels, provider as ProviderModels
from limes_common.models.elab import Sample

class Limes:
    def __init__(self) -> None:
        self._server = server.ServerConnection()
        self._eLab = eLab.ELabConnection()

        self.CallProvider = self._server.CallProvider
        self.GetSample = self._eLab.GetSample
        self.GetStorage = self._eLab.GetStorage
        self.GetFullStoragePath = self._eLab.GetFullStoragePath
        self.ReloadStorages = self._eLab.ReloadStorages

    #   def _auth() -> tuple[bool, str]:
    def _auth(self) -> bool:
        res = self._server.Authenticate()
        if res.Code != 200:
            return False

        if res.Success:
            print('Authenticated terminal as %s' % res.FirstName)
            # return True, res.FirstName
            self._eLab.SetToken(res.Token)
        else:
            # return False, ''
            print('Not logged in')
        return res.Success

    def Login(self, username, password) -> tuple[bool, str]:
        if not self._server.Ready and not self._server.Reconnect(): 
            return False, 'unable to connect to server'
        res = self._eLab.Login(username, password)

        if res.Code == 200:
            self._server.Login(res.user.firstName, res.user.lastName, res.token)
            return True, 'logged in as %s' % res.user.firstName
        elif res.Code == 401:
            return False, 'incorrect credentials'
        else:
            return False, 'failed with code [%s], msg [%s]' % (res.Code, res.Error)

    def Login_from_file(self, path: str) -> tuple[bool, str]:
        try:
            with open(path) as credFile:
                cred = credFile.readlines()
                u = cred[0]
                p = cred[1]
                return self.Login(u, p)
        except Exception as e:
            return False, str(e)

    def ListProviders(self):
        res = self._server.List()
        es = self._eLab.GetSchema()
        ei = ServerModels.ProviderInfo()
        ei.Name = 'eLab'
        ei.Schema = es
        if res.Providers is None:
            res.Providers = [ei]
        else:
            res.Providers.append(ei)
        return res

    def Search(self, query: str):
        s_res = self._server.Search(query)
        e_res = self._eLab.Search(query)
        return s_res.AddFrom(e_res)


from __future__ import annotations
import sys, subprocess
from typing import Callable
from threading import Condition, Thread
from queue import Queue, Empty
import json
import uuid
import traceback

from limes_common.utils import current_time
from . import Connection
from limes_common import config
from limes_common.models import Model, Primitive, provider as Models, ssh

_QUOTE = '\\q'
def Serialize(model: Model) -> str:
    return Console_Encode(json.dumps(model.ToDict()))
def Console_Encode(serial: str) -> str:
    return '"%s"' % (serial.replace('"', _QUOTE))
def Console_Decode(serial: str) -> str:
    return serial.replace(_QUOTE, '"')
# M = TypeVar('M')
# def Console_Decode(string: str, constr: Callable[[str], M]) -> M:
    # return constr(string.replace(_QUOTE, '"'))

class MessageID:
    def __init__(self, uuid: str = uuid.uuid4().hex) -> None:
        self.__uuid = uuid

    def AsHex(self):
        return self.__uuid
        
class SshConnection(Connection):
    class _Pipe:
        def __init__(self, io:subprocess.IO[bytes], lock: Condition=Condition(), q: Queue=Queue()) -> None:
            self.IO = io
            self.Lock = lock
            self.Q = q

    class _SshConsole:
        def __init__(self, console: subprocess.Popen[bytes], onOut: Callable[[str], None], onErr: Callable[[str], None],
                timeout: float) -> None:
            ios = {}
            for name, io in [('in', console.stdin), ('out', console.stdout), ('err', console.stderr)]:
                if io is None:
                    raise ValueError('an io is missing')
                else:
                    ios[name] = io
            self._in = SshConnection._Pipe(ios['in'])
            self._out = SshConnection._Pipe(ios['out'])
            self._err = SshConnection._Pipe(ios['err'])
            self._onCloseLock = Condition()
            self._closed = False

            workers: list[Thread] = []
            def reader(pipe: SshConnection._Pipe, callback: Callable[[str], None]):
                io = iter(pipe.IO.readline, b'')
                while True:
                    try:
                        line = next(io)
                    except (StopIteration, ValueError):
                        break
                    chunk = bytes.decode(line, encoding=config.ENCODING)
                    callback(chunk)
                self.Dispose()
            workers.append(Thread(target=reader, args=[self._out, onOut]))
            workers.append(Thread(target=reader, args=[self._err, onErr]))

            def timeoutTask():
                self._onCloseLock.acquire()
                self._onCloseLock.wait(timeout)
                self._onCloseLock.release()
                self.Dispose()
            workers.append(Thread(target=timeoutTask))
                
            for w in workers:
                w.daemon = True # stop with program
                w.start()

        def IsClosed(self) -> bool:
            self._onCloseLock.acquire()
            closed = self._closed
            self._onCloseLock.release()
            return closed

        def __write(self, pipe: SshConnection._Pipe, statement: str):
            # msg = ' '.join([cmd] + [Console_Encode(arg) for arg in args])
            pipe.IO.write(bytes('%s\n' % (statement), encoding=config.ENCODING))

        def Send(self, statement:str):
            stdin = self._in
            stdin.Lock.acquire()
            self.__write(stdin, statement)
            stdin.IO.flush()
            stdin.Lock.release()

        def BatchSend(self, statements: list[str]):
            stdin = self._in
            stdin.Lock.acquire()
            for s in statements:
                self.__write(stdin, s)
            stdin.IO.flush()
            stdin.Lock.release()

        def Dispose(self):
            self._onCloseLock.acquire()
            if self._closed:
                self._onCloseLock.release()
                return
            self._closed = True
            self._onCloseLock.notify_all()
            self._onCloseLock.release()
            try:
                self.Send('logout')
            except BrokenPipeError:
                pass
            for p in [self._in, self._out, self._err]:
                p.Lock.acquire()
                try:
                    p.IO.close()
                except BrokenPipeError:
                    pass
                p.Lock.release()

    class _Transaction:
        def __init__(self) -> None:
            self.Lock = Condition()
            self.Packets: Queue[str] = Queue()
            self.Errors: Queue[str] = Queue()
        
        def _sync(self, fn):
            self.Lock.acquire()
            fn()
            self.Lock.release()

        def Notify(self):
            self._sync(lambda: self.Lock.notify_all())
        
        def Wait(self, timeout: float):
            self._sync(lambda: self.Lock.wait(timeout))

    def __init__(self,
            url:str, setup: list[str], cmd: str,
            transactionTimeout:float, keepAliveTime:float,
            identityFile: str = None) -> None:
        super().__init__()
        self._transactions: dict[str, SshConnection._Transaction] = {}
        self._transactionTimeout = transactionTimeout
        self._keepAliveTime = keepAliveTime
        self._url = url
        self._identityFile = identityFile
        self._setup = setup
        self._cmd = cmd
        self.__connection = self._getCon()

        self._onResponseSubscribers = []
        self._onErrorSubScribers = []

    def _getCon(self) -> _SshConsole:
        cmd = ['ssh', '-tt', self._url]
        if self._identityFile is not None:
            cmd += ['-i', self._identityFile]
        # print(cmd)
        p = subprocess.Popen(cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
                    # time.sleep(2)
        setup = self._setup
        
        def onOut(msg):
            for cb in self._onResponseSubscribers:
                cb(msg)

            # if self.x == 0: print('>%s' % msg)
            if msg.startswith(Handler.SEND_FLAG):
                msg = msg[len(Handler.SEND_FLAG):]
                parsed = ssh.Message.Parse(msg)
                if parsed.MessageID in self._transactions:
                    tr = self._transactions[parsed.MessageID]
                    if parsed.IsError:
                        print('provider error: %s' % parsed.Body)
                        tr.Notify()
                    else:
                        tr.Packets.put(parsed.Body)
                        tr.Notify()
                else:
                    # print(parsed.MessageID)
                    # print([self._transactions.keys()])
                    print('unregistered message: %s'%msg)

        def onErr(msg):
            for cb in self._onErrorSubScribers:
                cb(msg)
            # if self.x == 0: print('er>%s' % msg)
            if not msg.startswith('Connection to') and not msg.endswith('closed.'):
                print('Fatal error> ' + msg)

        con = SshConnection._SshConsole(p, onOut, onErr, self._keepAliveTime)
        con.BatchSend(setup)
        return con

    def _send(self, ep:str, model: Model|None = None) -> MessageID:
        if self.__connection.IsClosed():
            self.__connection = self._getCon()

        cmd = self._cmd
        mid = MessageID()
        ser = Console_Encode(json.dumps(model.ToDict())) if model is not None else ''
        statement = '%s %s %s %s' % (cmd, mid.AsHex(), ep, ser)
        self.__connection.Send(statement)
        return mid

    def _listenFor(self, mid: MessageID, timeout:float=0) -> tuple[bool, str]:
        if timeout==0:
            timeout = self._transactionTimeout
        tr = SshConnection._Transaction()
        self._transactions[mid.AsHex()] = tr
        tr.Wait(timeout)
        try:
            return True, tr.Packets.get_nowait()
        except Empty:
            return False, ''

    def _makeTransaction(self, ep: str, body: Model|None = None) -> tuple[bool, str]:
        mid = self._send(ep, body)
        return self._listenFor(mid)

    def AddOnResponseCallback(self, fn: Callable[[str], None]) -> None:
        self._onResponseSubscribers.append(fn)

    def AddOnErrorCallback(self, fn: Callable[[str], None]) -> None:
        self._onErrorSubScribers.append(fn)

    def RemoveOnResponseCallback(self, fn: Callable[[str], None]) -> None:
        self._onResponseSubscribers.remove(fn)

    def RemoveOnErrorCallback(self, fn: Callable[[str], None]) -> None:
        self._onResponseSubscribers.remove(fn)

    # def CheckStatus(self, echo: str) -> Models.Status.Response:
    #     success, res = self._makeTransaction(ProviderEndpoint.CHECK_STATUS, Models.Status.Request(echo))
    #     if success:
    #         return Models.Status.Response.Parse(res)
    #     else:
    #         return Models.Status.Response(False, res)

    def GetSchema(self) -> Models.Schema:
        success, res = self._makeTransaction(Models.Endpoints.GET_SCHEMA)
        if success:
            return Models.Schema.Parse(res)
        else:
            return Models.Schema()

    def MakeRequest(self, request: Models.ProviderRequest) -> Models.GenericResponse:
        success, res = self._makeTransaction(request.TargetEndpoint, request)
        if success:
            resModel = Models.GenericResponse.Parse(res)
            return resModel
        else:
            err = Models.GenericResponse()
            err.Code = 503
            err.Error = 'connection failed'
            return err

    # todo: remove redundancy with make request
    def Search(self, query: str) -> Models.Search.Response:
        req = Models.Search.Request()
        req.Query = query
        success, res = self._makeTransaction(Models.Endpoints.SEARCH, req)
        if success:
            resModel = Models.Search.Response.Parse(res)
            return resModel
        else:
            err = Models.Search.Response()
            err.Code = 503
            err.Error = 'connection failed'
            return err

    def Dispose(self):
        self.__connection.Dispose()

class BadRequestException(Exception):
    pass

class Handler:
    SEND_FLAG = 'From Handler > '
    def __init__(self) -> None:
        self._lastRawRequest = ''

    def HandleCommandLineRequest(self) -> None:
        self._lastRawRequest = sys.argv
        if len(sys.argv) == 4:
            rmid, endpoint, body = sys.argv[1:]
        else:
            rmid, endpoint = sys.argv[1:]
            body = '{}'
        mid = MessageID(rmid)
        body = Console_Decode(body)

        def getHandlerMethod():
            PREFIX = '_parse'
            SUFFIX = 'Request'
            getMethod = lambda path: '%s_%s_%s' % (PREFIX, path, SUFFIX)
            # self._send(MessageID(), endpoint, True)
            for ep in Models.Endpoints.Paths():
                candidate = ep
                if candidate != endpoint: continue
                method = getMethod(candidate.title())
                myMethods = [m for m in dir(self) if m.startswith(PREFIX)]
                for m in myMethods:
                    if method == m:
                        return getattr(self, m)
                return method
            return self._parse_Generic_Request
        handler = getHandlerMethod()

        if isinstance(handler, str):
            self._send(mid, 'expectd method [%s] for endpoint [%s] not implimented' % (handler, endpoint), True)
        else:
            try:
                res = handler(endpoint, body)
                self._send(mid, json.dumps(res.ToDict()))
            except Exception as e:
                self._send(mid, '\n%sendpoint: %s'%(str(traceback.format_exc()), endpoint), True)
        pass

    def _send(self, mid: MessageID, msg: str, isError:bool = False):
        serialized = json.dumps(ssh.Message(mid.AsHex(), msg, isError).ToDict())
        print(Handler.SEND_FLAG + serialized)

    # def _parseStatusRequest(self, raw: str):
    #     return self.OnStatusRequest(Models.Status.Request.Parse(raw))
    # def OnStatusRequest(self, req: Models.Status.Request) -> Models.Status.Response:
    #     return Models.Status.Response(False, req.Msg, 'This is an abstract Provider and must be implemented')

    def _parse_Schema_Request(self, ep:str, raw: str) -> Models.Schema:
        return self.On_Schema_Request()
    def On_Schema_Request(self) -> Models.Schema:
        example = Models.Service()
        example.Endpoint = 'Abstract_service_example'
        example.Input = {'a': str, 'b': bool}
        example.Output = {'x': int}
        sch = Models.Schema()
        sch.Services = [example]
        return sch

    def _parse_Generic_Request(self, ep: str, raw: str) -> Models.GenericResponse:
        req = Models.GenericRequest.Parse(raw) # todo: dedicated primitive model
        try:
            if ep == '':
                res = Models.GenericResponse()
                res.Code = 400
                res.Error = 'Endpoint can not be empty'
                return res
            return self.On_Generic_Request(ep, req.Body)
        except Exception as e:
            return Models.GenericResponse({}, 500, str(e))
        
    def On_Generic_Request(self, endpoint: str, body: Primitive) -> Models.GenericResponse:
        """
        @purpose: name of service being invoked
        @data: data dict expected to follow the schema described by Service.Input
        @return: performs service listed by OnSchemaRequest() and return in form of Service.Output
        """
        res = Models.GenericResponse()
        res.Code = 500
        res.Error = 'provider not implemented!'
        return res

    def _parse_Search_Request(self, ep: str, raw: str) -> Models.Search.Response:
        req = Models.Search.Request.Parse(raw)
        try:
            return self.On_Search_Request(req.Query)
        except Exception as e:
            res = Models.Search.Response()
            res.Code = 500
            res.Error = str(e)
            return res

    def On_Search_Request(self, query: str) -> Models.Search.Response:
        res = Models.Search.Response()
        res.Code = 500
        res.Error = 'search not implemented!'
        return res



# todo: don't hard code this
version = \
'limes version 0.4.dev3'
help = '\n\n\
%s\n\
\n\
To run limes, one of the following is required:\n\
\n\
-h, --help\n\
    use with limes [command] -h to get sub options, example: limes login -h\n\
\n\
login\n\
    login to eLab using existing credentials\n\
    you will be prompted for a username and password\n\
\n\
get\n\
    *todo\n\
\n\
add\n\
    *todo\n\
' % version


from __future__ import annotations

from limes_common.models import Model, Primitive, provider as Models
from limes_common.models import provider
from limes_common.models.provider import Transaction
from limes_common.models.http import GET, POST

class Endpoints(provider.Endpoints):
    LOGIN = 'login'
    AUTHENTICATE = 'authenticate'
    INIT = 'init'
    # ADD = 'add'
    LIST = 'list'
    CALL = 'call'
    SEARCH = 'search'
    RESET = 'reset'

class ServerRequest(Models.ProviderRequest):
    ClientId: str

class ServerResponse(Models.ProviderResponse):
    pass

class Init(Transaction):
    class Request(ServerRequest):
        def __init__(self) -> None:
            super().__init__(Endpoints.INIT, GET)
    class Response(ServerResponse):
        CsrfToken: str
        def __init__(self, token: str = '') -> None:
            self.CsrfToken = token

class Authenticate(Transaction):
    class Request(ServerRequest):
        def __init__(self) -> None:
            super().__init__(Endpoints.AUTHENTICATE, POST)

    class Response(ServerResponse):
        Success: bool
        Token: str
        FirstName: str
        LastName: str

class Login(Transaction):
    class Request(ServerRequest):
        ELabKey: str
        FirstName: str
        LastName: str
        def __init__(self) -> None:
            super().__init__(Endpoints.LOGIN, POST)

    class Response(ServerResponse):
        Success: bool
        def __init__(self, success: bool=False) -> None:
            super().__init__()
            self.Success = success

class ProviderInfo(Model):
    Name: str
    Schema: provider.Schema

class List(Transaction):
    class Request(ServerRequest):
        def __init__(self) -> None:
            super().__init__(Endpoints.LIST, POST)

    class Response(ServerResponse):
        Providers: list[ProviderInfo]
        def __init__(self) -> None:
            super().__init__()
            self.Providers = []

class Reset(Transaction):
    class Request(ServerRequest):
        def __init__(self) -> None:
            super().__init__(Endpoints.RESET, POST)

    class Response(List.Response):
        pass

class CallProvider(Transaction):
    class Request(ServerRequest):
        ProviderName: str
        Endpoint: str
        RequestPayload: Models.GenericRequest
        def __init__(self) -> None:
            super().__init__(Endpoints.CALL, POST)

    class Response(ServerResponse):
        ResponsePayload: Models.GenericResponse

class Search(Transaction):
    class Request(ServerRequest, Models.Search.Request):
        def __init__(self) -> None:
            super().__init__(Endpoints.SEARCH, POST)

    class Response(Models.Search.Response):
        pass


from __future__ import annotations
from typing import Union

from requests.models import codes

from . import Model, Primitive
    
class Endpoints:
    GET_SCHEMA = 'schema'
    GENERIC = 'generic'
    SEARCH = 'search'

    @classmethod
    def Paths(cls):
        return [i[1] for i in filter(lambda i: i[0].isupper(), cls.__dict__.items())]

class ProviderRequest(Model):
    TargetEndpoint: str
    Method: str
    QueryParams: dict

    def __init__(self, endpoint: str='', method: str='', queryParams:dict={}) -> None:
        super().__init__()
        self.TargetEndpoint = endpoint
        self.Method = method
        self.QueryParams = queryParams

class GenericRequest(ProviderRequest):
    Body: Primitive

    def __init__(self, endpoint: str=Endpoints.GENERIC, method: str='', body: Primitive={}) -> None:
        super().__init__()
        self.Body = body
        self.TargetEndpoint = endpoint
        self.Method = method

class ProviderResponse(Model):
    Code: int
    Error: str
    Type: str

    def ToDict(self) -> dict:
        self.Type = self.GetType()
        return super().ToDict() 

    @classmethod
    def GetType(cls):
        return str(cls)

class GenericResponse(ProviderResponse):
    Body: Primitive

    def __init__(self, body: Primitive={}, code: int=200, error: str='') -> None:
        super().__init__()
        self.Body = body
        self.Code = code
        self.Error = error

    def BodyAsDict(self) -> dict:
        if isinstance(self.Body, dict):
            return self.Body
        else:
            return {'value': self.Body}

class Transaction:
    class Request(GenericRequest):
        pass

    class Response(GenericResponse):
        pass

DataSchema = Union[str, dict]
class Service(Model):
    Endpoint: str
    Input: DataSchema
    Output: DataSchema
    def __init__(self, endpoint: str='', input: DataSchema={}, output: DataSchema={}) -> None:
        super().__init__()
        self.Endpoint = endpoint
        self.Input = input
        self.Output = output

class Schema(ProviderResponse):
    Services: list[Service]
    def __init__(self, services: list[Service]=[]) -> None:
        super().__init__()
        self.Services = services

class Search(Transaction):
    class Request(ProviderRequest):
        Query: str
        Parameters: Primitive

    class Hit(Model):
        DataType: str
        Data: list[dict]

    class Response(ProviderResponse):
        Hits: dict[str, Search.Hit]
        def AddFrom(self, other: Search.Response):
            if self.Hits is None and 'Hits' in other.__dict__:
                self.Hits = other.Hits
            else:
                for k, v in other.Hits.items():
                    i = 2
                    key = k
                    while key in self.Hits:
                        key = '%s_%s' % (k, i)
                        i+=1
                    self.Hits[key] = v

            if self.Code == 200:
                self.Code = other.Code
                if 'Error' in other.__dict__:
                    self.Error = other.Error
            return self


import numpy as np

from limes_common import config
from limes_common.connections import T
from limes_common.connections.http import HttpConnection
from limes_common.models import Model, elab as Models, provider as ProviderModels

class Storages:
    def __init__(self, storages: list[Models.Storage] = None) -> None:
        if storages is None:
            try:
                loaded = np.load(file=config.ELAB_CACHE, allow_pickle=True)
                storages = [Models.Storage.Parse(s) for s in loaded]
            except FileNotFoundError:
                storages = []
        else:
            np.save(file=config.ELAB_CACHE, arr=[s.ToDict() for s in storages])

        S = Models.Storage
        self._byName: dict[str, list[S]] = {}
        self._byParentId: dict[int, list[S]] = {}
        self._byStorageId: dict[int, S] = {}
        self._byLayerId: dict[int, S] = {}

        def addToList(k, v, d:dict):
            similarLayers = d.get(k, [])
            similarLayers.append(v)
            d[k] = similarLayers

        for item in storages:
            item.link = ' %smembers/inventory/browser/?initStorageLayer=%s ' % (config.ELAB_URL, item.storageLayerID)
            addToList(item.name.lower(), item, self._byName)
            addToList(item.parentStorageLayerID, item, self._byParentId)
            self._byStorageId[item.storageID] = item
            self._byLayerId[item.storageLayerID] = item

class ELabConnection(HttpConnection):
    def __init__(self) -> None:
        super().__init__(config.ELAB_API)
        self._token: str = ''
        self._storages = Storages()

    def _makeHeader(self):
        return {'authorization': self._token}

    def SetToken(self, token:str) -> None:
        self._token = token

    def LoggedIn(self) -> bool:
        return self._token != ''

    def GetSchema(self) -> ProviderModels.Schema:
        s = ProviderModels.Schema()
        s.Code = 503
        s.Error = 'please visit https://elab.msl.ubc.ca/docs/api/ for eLab api documentation'
        return s

    def Search(self, query: str) -> ProviderModels.Search.Response:
        samples = self.SearchSamples(query)
        storages = self.SearchStorages(query)
        res = ProviderModels.Search.Response()
        t_samples = str(Models.Sample)
        t_storages = str(Models.Storage)
        def makeHit(x, t):
            hit = ProviderModels.Search.Hit()
            hit.DataType = t
            if x is None: return hit
            hit.Data = [i.ToDict() for i in x]
            return hit
        res.Hits = {
            'samples': makeHit(samples.data, t_samples),
            'storages': makeHit(storages, t_storages)
        }
        res.Code = 200
        return res

    def Login(self, username: str, password: str):
        transaction = Models.Login
        req = transaction.Request()
        req.username = username
        req.password = password
        res = self._makeParseRequest(
            req,
            transaction.Response.Parse,
            transaction.Response()
        )
        if res.Code == 200:
            self.SetToken(res.token)
        return res 

    def SearchSamples(self, query: str):
        query = query.lower()
        transaction = Models.SampleSearch
        req = transaction.Request(query)
        return self._makeParseRequest(
            req,
            transaction.Response.Parse,
            transaction.Response()
        )

    def GetSample(self, sampleID: int):
        transaction = Models.GetSample
        req = transaction.Request(sampleID)
        return self._makeParseRequest(
            req,
            transaction.Response.Parse,
            transaction.Response()
        )

    def ReloadStorages(self):
        transaction = Models.AllStorages
        req = transaction.Request()
        res = self._makeParseRequest(
            req,
            transaction.Response.Parse,
            transaction.Response()
        )
        self._storages = Storages(res.data)
        return res

    def SearchStorages(self, query: str):
        query = query.lower()
        results: list[Models.Storage] = []
        d = self._storages._byName
        for n in d.keys():
            if query in n:
                results += d[n]
        return results
    
    def GetStorage(self, layerID: int):
        # default = Models.Storage()
        default = None
        return self._storages._byLayerId.get(layerID, default)

    def GetFullStoragePath(self, layerID: int) -> list[Models.Storage]:
        layer = self.GetStorage(layerID)
        path = []
        while layer is not None:
            path.append(layer)
            layer = self.GetStorage(layer.parentStorageLayerID)
        return path

    # def _truncateToId(self, id: str) -> str:
    #     return id[-9:] if len(id) > 9 else id

    # def SearchSamplesById(self, strIds: list[str]) -> elab.Sample.ListResponse:
    #     ids = list(self._truncateToId(id) for id in strIds)
    #     query = '' if len(strIds)==0 else '?sampleID=' + reduce(lambda s, id: '%s,%s' % (s, id), ids, '')[1:]
    #     return elab.Sample.ListResponse(self.session.get(
    #         '%s/%s%s' % (self._makeUrl(ELabEndpoint.SAMPLES.Path), 'get', query),
    #         headers=self._getAuthHeader()
    #     ))

    # def SearchSamplesByName(self, name: str) -> elab.Sample.ListResponse:
    #     query = '' if len(name)==0 else '?name=' + reduce(lambda s, id: '%s,%s' % (s, id), [name], '')[1:]
    #     return elab.Sample.ListResponse(self.session.get(
    #         '%s%s' % (self._makeUrl(ELabEndpoint.SAMPLES.Path), query),
    #         headers=self._getAuthHeader()
    #     ))

    # def SearchSamples(self, token: str) -> elab.Sample.ListResponse:
    #     return elab.Sample.ListResponse(self.session.get(
    #         '%s%s%s' % (self._makeUrl(ELabEndpoint.SAMPLES.Path), '?search=', token),
    #         headers=self._getAuthHeader()
    #     ))

    # def GetSample(self, id: str) -> elab.Sample.Response:
    #     id = self._truncateToId(id)
    #     return elab.Sample.Response(self.session.get(
    #         '%s/%s' % (self._makeUrl(ELabEndpoint.SAMPLES.Path), id),
    #         headers=self._getAuthHeader()
    #     ))

    # def GetSampleMeta(self, id: str) -> elab.SampleMeta.Response:
    #     id = self._truncateToId(id)
    #     return elab.SampleMeta.Response(self.session.get(
    #         '%s/%s/meta' % (self._makeUrl(ELabEndpoint.SAMPLES.Path), id),
    #         headers=self._getAuthHeader()
    #     ))

    # def UpdateSampleMeta(self, sampleId: str, metaKey: str, newValue: str) -> elab.SampleMeta.UpdateResponse:
    #     id = self._truncateToId(sampleId)
    #     meta = self.GetSampleMeta(id)
    #     fieldId=0
    #     field = elab.MetaField()
    #     for k, f in meta.Fields.items():
    #         if k == metaKey:
    #             fieldId = f.sampleMetaID
    #             field = f
    #             if field.sampleDataType in ['TEXTAREA']:
    #                 field.value = newValue
    #             else:
    #                 print('cannot update a non text meta field')
    #     return elab.SampleMeta.UpdateResponse(self.session.patch(
    #         '%s/%s/meta/%s' % (self._makeUrl(ELabEndpoint.SAMPLES.Path), id, fieldId),
    #         headers=self._getAuthHeader(),
    #         data=field.ToDict(),
    #     ))


from typing import Union
from limes_common import config
from limes_common.models import Model, provider as Models
from limes_common.models.provider import GenericResponse, Transaction
from limes_common.models.http import GET, POST, PUT

class Endpoints(Models.Endpoints):
    AUTH = 'auth/user'
    SAMPLES = 'samples'
    STORAGE_LAYERS = 'storageLayers'

class ELabRequest(Models.ProviderRequest):
    pass

class ELabResponse(Models.ProviderResponse):
    pass

class User(Model):
    userID: int
    username: str
    email: str
    firstName: str
    lastName: str
    instituteID: int
    primaryGroupID: int
    primarySubGroupID: int
    domain: str
    
class Login(Transaction):
    class Request(ELabRequest):
        username: str
        password: str

        def __init__(self) -> None:
            super().__init__(Endpoints.AUTH, POST)

    class Response(ELabResponse):
        token: str
        user: User

class Authenticate(Transaction):
    class Request(ELabRequest):
        def __init__(self) -> None:
            super().__init__(Endpoints.AUTH, GET)

    class Response(ELabResponse):
        pass

class SampleType(Model):
    sampleTypeID: int
    userID: int
    groupID: int
    name: str

class Sample(Model):
    owner: str
    archived: bool
    sampleID: int
    created: str
    userID: int
    creatorID: int
    storageLayerID: int
    position: int
    seriesID: int
    groupID: int
    sampleTypeID: int
    sampleType: SampleType
    checkedOut: bool
    parentSampleID: int
    name: str
    description: str
    note: str
    link: str

    @classmethod
    def Parse(cls, raw: Union[bytes, str, dict], base: Model=None):
        s = super().Parse(raw, base=base)
        s.link = ' %smembers/inventory/browser/?return=true&sampleID=%s#viewSample ' % (config.ELAB_URL, s.sampleID)
        return s

class SearchHits(ELabResponse):
    recordCount: int
    currentPage: int
    maxRecords: int
    totalRecords: int
    data: list

class SampleSearch(Transaction):
    class Request(ELabRequest):
        def __init__(self, query: str='') -> None:
            super().__init__(Endpoints.SAMPLES, GET, queryParams={'search': query})

    class Response(SearchHits):
        data: list[Sample]

class SampleNameSearch(Transaction):
    class Request(ELabRequest):
        def __init__(self, query: str='') -> None:
            super().__init__(Endpoints.SAMPLES, GET, queryParams={'name': query})

    class Response(SearchHits):
        data: list[Sample]

class GetSample(Transaction):
    class Request(ELabRequest):
        def __init__(self, sampleID: int=0) -> None:
            super().__init__('%s/%s' % (Endpoints.SAMPLES, sampleID), GET)

    class Response(Sample, ELabResponse):
        pass

class Storage(Model):
    storageID: int
    storageLayerID: int
    parentStorageLayerID: int
    storageLayerDefinitionID: int
    position: int
    maxSize: int
    dimension: dict
    barcode: str
    isGrid: bool
    userID: int
    created: str
    name: str
    link: str

class AllStorages(Transaction):
    class Request(ELabRequest):
        def __init__(self) -> None:
            super().__init__(Endpoints.STORAGE_LAYERS, GET)

    class Response(SearchHits):
        data: list[Storage]

# class SampleModel:
#     def __init__(self, data: dict = {}) -> None:
#         self.Name = _tryParse(str, data, 'name', 'None')
#         self.Owner =_tryParse(str, data, 'owner', '')
#         self.TimeCreated =_tryParse(str, data, 'created', '')
#         self.Id =_tryParse(int, data, 'sampleID', 0)
#         self.Barcode =_tryParse(str, data, 'barcode', '')
#         self.Description =_tryParse(str, data, 'description', '')
#         self.ParentId =_tryParse(int, data, 'parent', 0)

# class Sample:
#     class ListResponse(PublicModel):
#         def __init__(self, res: Response) -> None:
#             super().__init__(res)
#             # self.Samples = list(map(lambda raw: SampleModel(raw), self.data))
#             self.Samples = list(map(lambda raw: SampleModel(raw), self.data.get('data', [])))

#     class Response(PublicModel):
#         def __init__(self, res: Response) -> None:
#             super().__init__(res)
#             self.Sample = SampleModel(self.data)


# class MetaField:
#     def __init__(self, data: dict = {}) -> None:
#         T = TypeVar('T')
#         def parse (constr, key):
#             defaults = {
#                 bool: False,
#                 str: '',
#                 int: 0,
#             }
#             return _tryParse(constr, data, key, defaults[constr])
#         self.raw = data
#         self.archived: bool = parse(bool, 'archived')
#         self.sampleMetaID: int = parse(int, 'sampleMetaID')
#         self.sampleDataType: str = parse(str, 'sampleDataType') # todo, make enum
#         self.key: str = parse(str, 'key')
#         self.value: str = parse(str, 'value')
#         self.sampleTypeMetaID: int = parse(int, 'sampleTypeMetaID')
#         self.truncated: bool = parse(bool, 'truncated')

#     def ToDict(self) -> dict:
#         out = self.__dict__.copy()
#         keys = list(out.keys())
#         for k in keys:
#             if k.startswith('_') or k in ['raw']:
#                 del out[k]
#         return out

# class SampleMeta:
#     class Response(PublicModel):
#         def __init__(self, res: Response) -> None:
#             super().__init__(res)
#             self.Fields: dict[str, MetaField] = {}
#             for raw in self.data['data']:
#                 self.Fields[raw['key']] = MetaField(raw)

#     class UpdateResponse(PublicModel):
#         def __init__(self, res: Response) -> None:
#             super().__init__(res)
#             self.Success = self.Code == 204



import requests as Requests
import uuid
from getpass import getuser
import os
from typing import Any, Callable, TypeVar, Union

from ..http import HttpConnection
from ... import config
from limes_common.models import Primitive, server as Models, provider as ProviderModels
# from limes_common.models.endpoints import ServerEndpoint

T = TypeVar('T')
class ServerConnection(HttpConnection):
    def __init__(self) -> None:
        super().__init__(config.SERVER_URL)
        self._id = ''
        self._csrf = ''
        self.Ready = False
        self.Reconnect()

    def Reconnect(self) -> bool:
        self._id = ('%012x:%s:%s' % (uuid.getnode(), getuser(), os.getppid()))
        self._csrf = ''
        req = Models.Init.Request()
        rawRes = self.MakeRequest(req, init=True)
        self.Ready = False
        if isinstance(rawRes.Body, dict):
            res = Models.Init.Response.Parse(rawRes.Body)
            if res.CsrfToken is not None and res.CsrfToken != '':
                self._csrf = res.CsrfToken
                self.Ready = True
        return self.Ready

    def _makeHeader(self):
        return {config.CSRF_NAME: self._csrf}

    def _makeJson(self, data: Models.ServerRequest) -> dict[str, Any]:
        data.ClientId = self._id
        d = data.ToDict()
        return d

    def MakeRequest(self, request: ProviderModels.ProviderRequest, init=False) -> ProviderModels.GenericResponse:
        def doErr():
            errmsg = 'failed to connect to limes server'
            return ProviderModels.GenericResponse(code=503, error=errmsg)

        if not self.Ready and not init:
            return doErr()

        try:
            return super().MakeRequest(request)
        except Requests.exceptions.ConnectionError:
            return doErr()
        except KeyboardInterrupt:
            return ProviderModels.GenericResponse(code=0, error='user interrupt')

    def Search(self, query: str):
        transaction = Models.Search
        req = transaction.Request()
        req.Query = query
        return self._makeParseRequest(
            req,
            transaction.Response.Parse,
            transaction.Response()
        )

    def Authenticate(self):
        transaction = Models.Authenticate
        return self._makeParseRequest(
            transaction.Request(),
            transaction.Response.Parse,
            transaction.Response()
        )

    def Login(self, firstName: str, lastName: str, token: str):
        req = Models.Login.Request()
        req.ELabKey = token
        req.FirstName = firstName
        req.LastName = lastName
        transaction = Models.Login
        return self._makeParseRequest(
            req,
            transaction.Response.Parse,
            transaction.Response()
        )

    def List(self):
        transaction = Models.List
        return self._makeParseRequest(
            transaction.Request(),
            transaction.Response.Parse,
            transaction.Response()
        )

    def ReloadProviders(self):
        transaction = Models.Reset
        return self._makeParseRequest(
            transaction.Request(),
            transaction.Response.Parse,
            transaction.Response()
        )

    def CallProvider(self, providerName: str, providerEndpoint: str, body: Primitive):
        transaction = Models.CallProvider
        req = transaction.Request()
        req.ProviderName = providerName
        payload = ProviderModels.GenericRequest()
        payload.TargetEndpoint = providerEndpoint
        payload.Body = body
        req.RequestPayload = payload
        return self._makeParseRequest(
            req,
            transaction.Response.Parse,
            transaction.Response()
        )



from datetime import datetime, timezone
from enum import Enum

# prevents the enum name from printed
# str rep changes from enum.key -> key
class AbbreviatedEnum(Enum):
    def __str__(self) -> str:
        default = super().__str__()
        start = default.find(self.name)
        return default[start:]

# allows for more values to be associated
# ex ENUM_KEY = a, b, c
#    override __init__ to accept the additional values (b, c)
# HTTPMethod is an example
class AdvancedEnum(Enum):
    def __new__(cls, *args, **kwds):
        obj = object.__new__(cls)
        obj._value_ = args[0]
        return obj

# str rep is UTC
class SerializableTime(datetime):
    def __str__(self) -> str:
        prev = self.tzinfo
        self.replace(tzinfo=timezone.utc)
        utc = self.timestamp()
        self.replace(tzinfo=prev)
        return str(utc)

    def __repr__(self) -> str:
        return self.__str__()

# has method to hide attributes with "_"
class PublicOnlyDict:
    def GetDict(self):
        public = {}
        for k, v in self.__dict__.items():
            if not k.startswith('_'):
                public[k] = v
        return public



from typing import Any, TypeVar, Union, Callable
from limes_common.models.basic import AbbreviatedEnum

from limes_common.models import Primitive, Model, provider as Models
from limes_common.exceptions import AbstractMethodException
from limes_common.utils import current_time

_notImplimentedMsg = 'Call on abstract provider. Funciton not implimented.'

T = TypeVar('T')
class Connection:
    # def CheckStatus(self, echo: str='') -> Models.Status.Response:
    #     raise AbstractClassException(_notImplimentedMsg)

    def GetSchema(self) -> Models.Schema:
        raise AbstractMethodException(_notImplimentedMsg)

    def MakeRequest(self, request: Models.ProviderRequest) -> Models.ProviderResponse:
        raise AbstractMethodException(_notImplimentedMsg)

    def Search(self, query: str) -> Models.Search.Response:
        raise AbstractMethodException(_notImplimentedMsg)

    def Close(self):
        pass



from __future__ import annotations
import inspect
from json.decoder import JSONDecodeError
from typing import Callable, TypeVar, Union, Any, get_type_hints
import json


# U = TypeVar('U')
Primitive = Union[str, int, float, bool, list['Primitive'], dict[str, 'Primitive']]

T = TypeVar('T')
TypeDesc = Union[str, tuple[str, list['TypeDesc']]]
class SerializableTypes:
    # all need to be strict uppercase for _members to see
    # constructor, type
    STR = str, str
    BOOL = lambda x: SerializableTypes._parseBool(x), bool
    INT = int, int
    FLOAT = lambda x: SerializableTypes._parseFloat(x), float
    PRIMITIVE = lambda x: SerializableTypes._parseCustomType(Primitive, x, False), 'Primitive'
    TYPE = lambda x: SerializableTypes._parseTypeAsString(x), type

    DefinedModels: dict[str, type[Model]] = {}

    @classmethod
    def _parseFloat(cls, x) -> float:
        if isinstance(x, float):
            return x
        f = float(x)
        if f.is_integer():
            raise ValueError
        else: 
            return f

    @classmethod
    def _parseBool(cls, x) -> bool:
        if isinstance(x, bool):
            return x
        elif isinstance(x, str):
            s = x.lower()
            if s == 'false':
                return False
            elif s == 'true':
                return True
        raise ValueError

    @classmethod
    def _parseCustomType(cls, customType: type, val, default: T) -> T:
        type = cls.ExtractTypes(str(customType))
        success, value = cls.Parse(type, val)
        if success:
            return value
        else:
            return default # default value

    @classmethod
    def _parseTypeAsString(cls, x) -> str:
        x = str(x)
        if x.startswith('<class'):
            return cls._typeStr(x)
        else:
            raise ValueError

    @classmethod
    def _typeStr(cls, t) -> str:
        s = str(t).lower()
        if s.startswith('<class'):
            return s[8:-2]
        if s.startswith("'") and s.endswith("'"):
            return s[1:-1]
        return s 

    __typesMap = {} # this may not be necessary if children to overwrite this
    @classmethod
    def _members(cls) -> list[dict[str, Callable[[str], Any]]]:
        def get(c: type):
            allMap = SerializableTypes.__typesMap
            m = allMap.get(c, None)
            if m == None:
                d = {}
                for name, items in [i for i in filter(lambda i: i[0].isupper(), c.__dict__.items())]:
                    # d[str(type(default))[8:-2]] = makeSafeConstr(constr, default)
                    constr, theType = items
                    d[cls._typeStr(theType)] = constr
                allMap[c] = d
                m = d
            return m
        all = []
        # print(inspect.getmro(cls))
        for c in inspect.getmro(cls):
            all.append(get(c))
            if c == SerializableTypes: break
        # print([[d.keys()] for d in all])
        return all

    @classmethod
    def _registerModel(cls, child):
        if str(child) == "<class 'limes_common.models.Model'>" or (inspect.isclass(child) and issubclass(child, Model)):
            name = cls._typeStr(child)
            if name not in cls.DefinedModels:
                cls.DefinedModels[name] = child

    @classmethod
    def ConstrFromStr(cls, string: str) -> Callable | None:
        string = cls._typeStr(string)
        for d in cls._members():
            # print([d.keys()])
            if string in d:
                return d[string]
        return None

    @classmethod
    def ConstrFromType(cls, type: type) -> Callable | None:
        string = str(type)
        return cls.ConstrFromStr(string)

    @classmethod
    def ExtractTypes(cls, hint:str, start:int = 0) -> TypeDesc:
        def extractTypes(start:int = 0) -> tuple[list[TypeDesc], int]:
            if hint.startswith('<class'):
                return [hint[8:-2]], start
            else:
                types = []
                token = ''
                def reset():
                    nonlocal token
                    token = ''

                i = start-1
                while i < len(hint) - 1:
                    i+=1
                    c = hint[i]
                    if c == ',':
                        types.append(token)
                        reset()
                        continue
                    elif c == '[':
                        subs, i = extractTypes(i+1)
                        types.append((token, subs))
                        reset()
                        continue
                    elif c == ']':
                        types.append(token)
                        return types, i+1
                    if c != ' ': token += c

                if len(token) > 0:
                    types.append(token)
                return types, i
        t, _ = extractTypes()
        return t[0]

    @classmethod
    def Parse(cls, type: TypeDesc, value, silent=False) -> tuple[bool, Any]:
        if value is None: return False, None

        if isinstance(type, tuple):
            base, subs = type
        else:
            base, subs = type, []

        if base == 'dict' and isinstance(value, dict):
            if len(subs) == 2:
                keyType, valType = subs
                d = {}
                success = True
                for k, v in value.items():
                    s1, key = cls.Parse(keyType, k)
                    s2, d[key] = cls.Parse(valType, v)
                    success = s1 and s2 and success
                return success, d
            else:
                return True, value

        if base == 'list' and isinstance(value, list):
            if len(subs) == 1:
                itemType = subs[0]
                l = []
                success = True
                for i in value:
                    s, v = cls.Parse(itemType, i)
                    l.append(v)
                    success = success and s
                return success, l
            else:
                return True, value

        if base == 'typing.Union':
            order = ['dict', 'tuple', 'list', 'bool', 'float', 'int', 'str']
            sorted: list[Union[None, TypeDesc]] = [None] * len(order)
            others = []
            success = False
            for t in subs:
                try:
                    i = order.index(str(t))
                    sorted[i] = t
                except ValueError:
                    others.append(t)
            for t in others + sorted:
                if t is None: continue
                s, parsed = cls.Parse(t, value, silent=True)
                if s: return True, parsed

        baseLower = base.lower()
        if base.lower() in cls.DefinedModels:
            theModel = cls.DefinedModels[baseLower]
            if isinstance(value, dict) or isinstance(value, bytes) or isinstance(value, str):
                return True, theModel.Parse(value)

        constr = cls.ConstrFromStr(base)
        if constr is not None:
            try:
                if base == 'type': print(base, value)
                return True, constr(value)
            except (TypeError, ValueError):
                pass

        if not silent:
            if len(subs) > 0:
                subPrint = []
                for s in subs:
                    if isinstance(s, tuple):
                        subPrint.append(s[0])
                    else:
                        subPrint.append(s)
            else:
                subPrint = ''
            print('defaulting to [%s]\nfor unknown type <%s%s>'%(value, base, subPrint))
            print('types inheriting Model must call the super-constructor if implimenting __init__')
            print('custom types must use a custom <SerializableTypes>')
        return False, value


class AutoRegisterChildren(type):
    def __new__(cls, name, bases, classdict):
        new_cls = type.__new__(cls, name, bases, classdict)
        SerializableTypes._registerModel(new_cls)
        return new_cls

class Model(metaclass=AutoRegisterChildren):
    GetTypesDict: Callable[..., type[SerializableTypes]] = lambda: SerializableTypes

    # INHERITING CLASSES MUST SET DEFAULTS TO ALL CONSTRUCTOR PARAMATERS
    def __init__(self) -> None:
        self._jsonLoadSuccess = False
        self._raw = ''

    @classmethod
    def _getTypes(cls, typesDict: type[SerializableTypes]) -> dict[str, TypeDesc]:
        hints = get_type_hints(cls)
        d = {}
        for k, hint in hints.items():
            hint = str(hint)
            type = typesDict.ExtractTypes(hint)
            d[k] = type
        return d

    @classmethod
    def Parse(cls, raw: bytes | str | dict, base: Model=None):
        try:
            model = cls()
        except TypeError:
            raise TypeError('%s inheriting <Model> must have no requred paramaters in constructor' % str(cls))

        try:
            if isinstance(raw, dict):
                d = raw
            else:
                d = json.loads(raw)
                model._jsonLoadSuccess = True
        except JSONDecodeError:
            model._raw = raw
            return model

        typesDict = cls.GetTypesDict()
        hints = cls._getTypes(typesDict)
        # print([hints.keys()])
        for k, t in hints.items():
            if k.startswith('_') or k == 'GetTypesDict': continue
            _, val = typesDict.Parse(t, d.get(k, None))
            if val is None and base is not None:
                val = base.__dict__.get(k, None)
            model.__setattr__(k, val)
        return model

    def _toDict(self, v):
        if isinstance(v, Model):
            return v.ToDict()
        elif isinstance(v, list):
            l = []
            for i in v:
                l.append(self._toDict(i))
            return l
        elif isinstance(v, dict):
            d = {}
            for k, vv in v.items():
                d[k] = self._toDict(vv)
            return d
        elif isinstance(v, type):
            return str(v)
        else:
            return v

    def ToDict(self) -> dict:
        d = {}
        for k, v in self.__dict__.items():
            if k.startswith('_'): continue
            d[k] = self._toDict(v)
        return d

# decorator to register class
def Serializable(cls: T) -> T:
    SerializableTypes._registerModel(cls)
    return cls


from limes_common.models import Model

class Message(Model):
    MessageID: str
    Body: str
    IsError: bool
    
    def __init__(self, mid: str='', body: str='', isError:bool=False) -> None:
        self.MessageID = mid
        self.Body = body
        self.IsError = isError